//
// Cuda Kernel
//
///////////////////////////////////////////////////////////////////////////////

#include "DFRKernelDefines.h"

#define __THREADS_X 16
#define __THREADS_Y 16

#define TRACE_DIST  512.0f
#define SHADOW_EPS    0.1f

// #define CALL_PER_Y

//
// Declare consts
//
__constant__ float4 gCameraPosition;
__constant__ float4 gProjectionVector;
__constant__ float4 gModelviewMatrix[4];
__constant__ uint4  gTextureDims;

///////////////////////////////////////////////////////////////////////////////
//
// Prototypes
//
extern "C" void CUDA_DFKernel(void        *Surface, 
							  unsigned int RTSizeX, 
							  unsigned int RTSizeY);

extern "C" void CUDA_BindConsts(unsigned int RTSizeX,
								unsigned int RTSizeY,
								unsigned int RTPitch,
								const float *RayOrigin,
								const float *ProjVector,
								const float *ModelviewInv);

///////////////////////////////////////////////////////////////////////////////
__device__ float3 TransformRayDir(const float3 &RayDirection)
{
 float3 res;

 res.x = (gModelviewMatrix[0].x * RayDirection.x) +
		 (gModelviewMatrix[1].x * RayDirection.y) +
		 (gModelviewMatrix[2].x * RayDirection.z) ;

 res.y = (gModelviewMatrix[0].y * RayDirection.x) +
		 (gModelviewMatrix[1].y * RayDirection.y) +
		 (gModelviewMatrix[2].y * RayDirection.z) ;

 res.z = (gModelviewMatrix[0].z * RayDirection.x) +
		 (gModelviewMatrix[1].z * RayDirection.y) +
		 (gModelviewMatrix[2].z * RayDirection.z) ;

 return res;
}
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
__device__ float DFPRT_StrangeObject(const float3 &Pt, float DSMultiply)
{
 float3 center = make_float3(0.0f,-32.0f, 0.0f) - Pt;
 float  ds     = sqrtf(dot(center, center));
        center*= __fdividef(1.0f, ds);
		ds    -= 16.0f;
		ds    += __sinf(center.y *-12.0f) * 4.0f;

 return fmaxf(0.0f, ds * DSMultiply);
}
///////////////////////////////////////////////////////////////////////////////
__device__ float DFPRT_Floor(const float3 &Pt)
{
 float3 n = make_float3(0.0f,  1.0f, 0.0f);
 float3 p = make_float3(0.0f,-20.0f, 0.0f);
 float  w = -dot(p, n);
 float  f = -dot(n, Pt) + w;

 return fabsf(f);
}
///////////////////////////////////////////////////////////////////////////////
__device__ float DFPRT_Tackles(const float3 &Pt, float DSMultiply)
{
 float d = TRACE_DIST;

 float dsort = dot(make_float3(Pt.x, Pt.z, 0.0f), make_float3(Pt.x, Pt.z, 0.0f));
 float dsrts = sqrtf(dsort);

 float rd    = fmaxf(0.0f, __cosf((__fdividef(fminf(333.0f, dsrts), 333.0f)) * M_PI) * 3.0f);
	   rd   *= 1.0f + (fabsf(__cosf(__fdividef(dsrts, 50.0f))) * 0.5f);

 float ox    = (__cosf(dsort * 0.0007f) * (__sinf(dsort * 0.0002f) * 8.0f));
 float oy    = (expf(dsort * -0.0025f) * 32.0f) - 16.0f;
 float fc    = (__sinf((dsrts * 0.0075f)) * (M_PI / 2.0f));

 #pragma unroll
 for (unsigned int i = 0; i < 8; i++)
 {
	 float3 P   = rotateY(Pt, (M_PI * 2.0f * __fdividef((float)i, 8.0f)) + fc);
			P.z = 0.0f;
			P.x+= ox;
			P.y+= oy;

	 float dt   = fmaxf(0.0f, sqrtf(dot(P, P)) - rd);

	 if (rd < MATH_EPSILON)
		 dt = TRACE_DIST;

	 d = fminf(dt * DSMultiply, d);
 }

 return d;
}
///////////////////////////////////////////////////////////////////////////////
__device__ float DistanceFunct(const float3 &ROrg)
{
 //
 // Distance function for casting (fixups for distance evaluations)
 //
 float CDst =			  DFPRT_StrangeObject(ROrg, 0.25f);
	   CDst = fminf(CDst, DFPRT_Floor(ROrg));
	   CDst = fminf(CDst, DFPRT_Tackles(ROrg, 0.5f));

 return CDst;
}
///////////////////////////////////////////////////////////////////////////////
__device__ float DistanceFunctAO(const float3 &ROrg)
{
 //
 // Distance function for AO & SoftShadows (no fixups)
 //
 float CDst =			  DFPRT_StrangeObject(ROrg, 1.0f);
	   CDst = fminf(CDst, DFPRT_Floor(ROrg));
	   CDst = fminf(CDst, DFPRT_Tackles(ROrg, 1.0f));

 return CDst;
}
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
__device__ float DistanceFunctCast(const float3 &RayOrigin, const float3 &RayDirection)
{
 float  TDst = 0.0f;
 float3 ROrg = RayOrigin;
 int    IMax = 128;
 do
 {
	 float CDst = DistanceFunct(ROrg);
	       TDst+= CDst;

	 if (CDst < MATH_EPSILON)
		 break;

	 ROrg += (RayDirection * CDst);

	 IMax--;
	 if (IMax == 0)
		 break;

 } while (TDst < TRACE_DIST);

 return TDst;
}
///////////////////////////////////////////////////////////////////////////////
__device__ float CalculateAO(const float3 &Pw, const float3 &Nn)
{
 //
 // Proximity information interpreted as AO
 //
 const float AOSample = 1.0f;
 const int   AOSteps  = 8;

 float AO = 0.0f;

 for (int i = 0; i < AOSteps; i++)
 {
	 float ds = (float)(i+1) * AOSample;
	 float px = DistanceFunctAO(Pw + (Nn * ds));

	 AO += __fdividef(fmaxf(0.0f, ds - px), ds);
 }

 AO *= (1.0f / (float)AOSteps);

 return 1.0f - fminf(1.0f, AO);
}
///////////////////////////////////////////////////////////////////////////////
__device__ float CalculateSoftShadow(const float3 &Pw, const float3 &Lv, float LDist)
{
 //
 // Soft shadows
 //
 float d = DistanceFunctCast(Pw + (Lv * SHADOW_EPS), Lv);

 if (d < LDist)
	 return 0.0f;

 //
 // TODO::Soft shadows -> trace distance to occluder along Lv and forge softshadow info from this
 //

 return 1.0f;
}
///////////////////////////////////////////////////////////////////////////////
__device__ float3 MakeLights(const float3 &Pw, const float3 &Nn, const float3 &CameraPosition)
{
 //
 // Light
 //
 float3 Vv = CameraPosition - Pw;
		Vv*= rsqrtf(dot(Vv, Vv));
 float  Df;

 Df = fmaxf(0.0f, dot(make_float3(-0.577f, -0.577f, -0.577f), Nn));
 if (Df > MATH_EPSILON)
	 Df *= CalculateSoftShadow(Pw, make_float3(-0.577f, -0.577f, -0.577f), TRACE_DIST);

 float3 LColor = make_float3(1.0f, 0.3f, 0.3f) * Df;

 float3 Hv = make_float3(-0.577f, -0.577f, -0.577f) + Vv;
		Hv*= rsqrtf(dot(Hv, Hv));

 LColor += make_float3(0.0f, 0.7f, 0.0f) * powf(fmaxf(0.0f, dot(Hv, Nn)), 16.0f);

 Df = fmaxf(0.0f, dot(make_float3(0.000f, -1.000f, 0.000f), Nn));
 if (Df > MATH_EPSILON)
	 Df *= CalculateSoftShadow(Pw, make_float3(0.000f, -1.000f, 0.000f), TRACE_DIST);

 LColor += make_float3(0.2f, 0.2f, 0.6f) * Df;

 Df = fmaxf(0.0f, dot(make_float3(0.577f, -0.577f, 0.577f), Nn));
 if (Df > MATH_EPSILON)
	 Df *= CalculateSoftShadow(Pw, make_float3(0.577f, -0.577f, 0.577f), TRACE_DIST);

 LColor += make_float3(0.4f, 0.4f, 0.4f) * Df;

 return LColor;
}
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
__global__ void DFKernel(unsigned char *Surface, unsigned int OffY)
{
 const int x = (blockIdx.x * blockDim.x) + threadIdx.x;
 const int y = (blockIdx.y * blockDim.y) + threadIdx.y + OffY;

 if (x >= gTextureDims.x || y >= gTextureDims.y) 
	 return;
	
 unsigned char *Pixel = Surface + (y * gTextureDims.z) + (x<<2);

 //
 // Create ray direction
 //
 float3 RayDirection;
		RayDirection.x = (((float)(x) * gProjectionVector.x) - 1.0f) * gProjectionVector.z;
		RayDirection.y = (((float)(y) * gProjectionVector.y) - 1.0f) * gProjectionVector.w;
		RayDirection.z = -1.0f;
		RayDirection   = TransformRayDir(RayDirection);
		RayDirection  *= rsqrtf(dot(RayDirection, RayDirection));

 float3 RayColor       = make_float3(0.3f, 0.4f, 0.3f);
 float3 RayOrigin      = _asflt3(gCameraPosition);

 //////////////////////////////////////////////////////////////////////////////
 //
 // Distance field renderer itself
 //
 float Distance = DistanceFunctCast(RayOrigin, RayDirection);

 if (Distance < TRACE_DIST)
 {
	 const float eps = 0.01f;
	 float3 Pw = RayOrigin + (RayDirection * Distance);
	 float3 Normal;

	 float3 PwN = Pw - (RayDirection * eps);

	 Normal.x = DistanceFunctAO(PwN + make_float3(eps,  0.0f, 0.0f)) - DistanceFunctAO(PwN - make_float3(eps,  0.0f, 0.0f));
	 Normal.y = DistanceFunctAO(PwN + make_float3(0.0f, eps,  0.0f)) - DistanceFunctAO(PwN - make_float3(0.0f, eps,  0.0f));
	 Normal.z = DistanceFunctAO(PwN + make_float3(0.0f, 0.0f, eps )) - DistanceFunctAO(PwN - make_float3(0.0f, 0.0f, eps ));

     Normal  *= rsqrtf(dot(Normal, Normal));

	 float AO = CalculateAO(Pw, Normal);

	 RayColor = MakeLights(Pw, Normal, RayOrigin) * AO;
 }

 //////////////////////////////////////////////////////////////////////////////

 Pixel[3] = 0xFF;						  // A
 Pixel[2] = __float_as_uchar(RayColor.x); // R
 Pixel[1] = __float_as_uchar(RayColor.y); // G
 Pixel[0] = __float_as_uchar(RayColor.z); // B

 return;
}
///////////////////////////////////////////////////////////////////////////////
//
// Caller
//
void CUDA_DFKernel(void        *Surface, 
				   unsigned int RTSizeX, 
				   unsigned int RTSizeY)
{
 unsigned int GridDimsX = ((unsigned int)RTSizeX + (__THREADS_X-1)) / __THREADS_X;
 unsigned int GridDimsY = ((unsigned int)RTSizeY + (__THREADS_Y-1)) / __THREADS_Y;

#ifndef CALL_PER_Y
 dim3 GridDimension  = dim3(GridDimsX, GridDimsY);
 dim3 BlockDimension = dim3(__THREADS_X, __THREADS_Y);

 DFKernel<<<GridDimension, BlockDimension>>>((unsigned char *)Surface, 0);

 cudaError_t err = cudaGetLastError();
 if (err != cudaSuccess)
	 fucked_error("Cuda failed to execute kernel with message: [%s]", cudaGetErrorString(err));
#else
 for (unsigned int y = 0; y < GridDimsY; y++)
 {
	 dim3 GridDimension  = dim3(GridDimsX, 1);
	 dim3 BlockDimension = dim3(__THREADS_X, __THREADS_Y);

	 __TMS_CU();
	 DFKernel<<<GridDimension, BlockDimension>>>((unsigned char *)Surface, y * __THREADS_Y);
	 __TME_CU("Pass");

	 cudaError_t err = cudaGetLastError();
	 if (err != cudaSuccess)
		 fucked_error("Cuda failed to execute kernel with message: [%s]", cudaGetErrorString(err));
 }
#endif

 return;
}
///////////////////////////////////////////////////////////////////////////////
void CUDA_BindConsts(unsigned int RTSizeX,
					 unsigned int RTSizeY,
					 unsigned int RTPitch,
					 const float *RayOrigin,
					 const float *ProjVector,
					 const float *ModelviewInv)
{
 if (cudaMemcpyToSymbol(gCameraPosition, RayOrigin, sizeof(float4)) != cudaSuccess)
	 fucked_error("Cannot bind Cuda symbol !");
 
 if (cudaMemcpyToSymbol(gProjectionVector, ProjVector, sizeof(float4)) != cudaSuccess)
	 fucked_error("Cannot bind Cuda symbol !");

 if (cudaMemcpyToSymbol(gModelviewMatrix, ModelviewInv, sizeof(float4)*4) != cudaSuccess)
	 fucked_error("Cannot bind Cuda symbol !");

 unsigned int TexDim[4] = 
 {
	 (unsigned int)RTSizeX, 
	 (unsigned int)RTSizeY,
	 (unsigned int)RTPitch,
	 (unsigned int)0
 };

 if (cudaMemcpyToSymbol(gTextureDims, TexDim, sizeof(uint4)) != cudaSuccess)
	 fucked_error("Cannot bind Cuda symbol !");

 // mxLogger("RORG: %f, %f, %f, %f", RayOrigin[0], RayOrigin[1], RayOrigin[2], RayOrigin[3]);
 // mxLogger("PVEC: %f, %f, %f, %f", ProjVector[0], ProjVector[1], ProjVector[2], ProjVector[3]);
 // mxLogger("MMAT: %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f, %f", ModelviewInv[0], ModelviewInv[1], ModelviewInv[2], ModelviewInv[3], ModelviewInv[4], ModelviewInv[5], ModelviewInv[6], ModelviewInv[7], ModelviewInv[8], ModelviewInv[9], ModelviewInv[10], ModelviewInv[11], ModelviewInv[12], ModelviewInv[13], ModelviewInv[14], ModelviewInv[15]);

 return;
}
///////////////////////////////////////////////////////////////////////////////
